@@ -32,10 +32,6 @@
 #include "main.h"
 #include "action.h"
 
-#ifndef _WIN32
-#include <pwd.h>
-#endif
-
 #include <synfig/general.h>
 
 #include <synfig/color.h>
@@ -331,7 +327,7 @@ synfigapp::Main::set_interpolation(synfig::Waypoint::Interpolation x)
 	{
 		interpolation_=x;
 
-		signal_interpolation_changed();
+		signal_interpolation_changed()();
 	}
 }
 
@@ -439,21 +435,10 @@ synfig::String
 synfigapp::Main::get_user_app_directory()
 {
 	String dir;
-	//! \todo do we need something like Glib::locale_from_utf8()?  (bug #1837445)
-#ifdef _WIN32
-	dir = String(getenv("HOMEDRIVE"))+getenv("HOMEPATH");
-#else
-	struct passwd* pwd = getpwuid(getuid());
-	if (pwd)
-	{
-		dir = pwd->pw_dir;
+	if (char* synfig_user_settings_dir = getenv("SYNFIG_USER_SETTINGS")) {
+		dir =  Glib::locale_from_utf8(String(synfig_user_settings_dir));
+	} else {
+		dir = Glib::get_home_dir()+ETL_DIRECTORY_SEPARATOR+SYNFIG_USER_APP_DIR;
 	}
-	else
-	{
-		// try the $HOME environment variable
-		dir = getenv("HOME");
-	}
-#endif
-	dir = dir+ETL_DIRECTORY_SEPARATOR+SYNFIG_USER_APP_DIR;
 	return dir;
 }
